// hlavickove subory *******************************************************
#include "../includes.h"
// vlastny .h subor
#include "fbo.h"

// konstruktor *************************************************************
CFBO::CFBO()
{
 created=-1; // FBO neexistuje
}

// destruktor **************************************************************
CFBO::~CFBO()
{
 cancel(); // zrusenie FBO

 glDeleteBuffers(1,&v_buf);
 glDeleteBuffers(1,&t_buf);

 // Vertex Array Object
 glDeleteVertexArrays(1,&vao);
}

// inicializacia stvoruholnika predstavujuceho FBO zlozeny z 2 trojuholnikov
void CFBO::init_obj(void) {
 // stvoruholnik predstavujuci FBO zlozeny z 2 trojuholnikov
 loopi(18) v[i]=0.0f;
 v[0]=-1.0f; v[1]=-1.0f; v[3]=1.0f; v[4]=-1.0f; v[6]=1.0f; v[7]=1.0f;
 v[9]=-1.0f; v[10]=-1.0f; v[12]=1.0f; v[13]=1.0f; v[15]=-1.0f; v[16]=1.0f;
 t[0]=0.0f; t[1]=0.0f; t[2]=1.0f; t[3]=0.0f; t[4]=1.0f; t[5]=1.0f;
 t[6]=0.0f; t[7]=0.0f; t[8]=1.0f; t[9]=1.0f; t[10]=0.0f; t[11]=1.0f;

 // Vertex Array Object
 glGenVertexArrays(1,&vao);
 glBindVertexArray(vao);

 glGenBuffers(1,&v_buf);
 glBindBuffer(GL_ARRAY_BUFFER,v_buf);
 glBufferData(GL_ARRAY_BUFFER,sizeof(v),v,GL_STATIC_DRAW);
 // vrcholy
 glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,0,0);
 glGenBuffers(1,&t_buf);
 glBindBuffer(GL_ARRAY_BUFFER,t_buf);
 glBufferData(GL_ARRAY_BUFFER,sizeof(t),t,GL_STATIC_DRAW);
 // texturovacie suradnice
 glVertexAttribPointer(8,2,GL_FLOAT,GL_FALSE,0,0);

 glEnableVertexAttribArray(0);
 glEnableVertexAttribArray(8);

 glBindVertexArray(0);
}

// inicializacia FBO0 ******************************************************
int CFBO::init0(void)
{
 // FBO0 pre scr_w, scr_h
 glGenFramebuffersEXT(1,&f0.id);
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f0.id);

 // 1 depth
 glGenRenderbuffersEXT(1,&f0.depth);
 glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,f0.depth);
 glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT,scr_w,scr_h);
 glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,f0.depth);
 glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,0);

 // 2 frame
 loopi(2) {
  glGenTextures(1,&(f0.tex_id[i]));
  glBindTexture(GL_TEXTURE_2D,f0.tex_id[i]);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,scr_w,scr_h,0,GL_RGBA,GL_FLOAT,NULL);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  switch (i) {
   case 0: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,f0.tex_id[i],0); break;
   case 1: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D,f0.tex_id[i],0); break;
  }

  printf("FBO0 frame textura: %u\n",f0.tex_id[i]);
 }

 ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
 if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
  printf("FBO0 error: 0x%x\n",ret); return(1);
 }

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

 init_obj(); // inicializacia stvoruholnika

 created=0; // FBO0 bol uspesne vytvoreny

 return(0);
}

// inicializacia FBO1 a FBO2 ***********************************************
int CFBO::init(void)
{
 // FBO1 pre scr_w, scr_h
 glGenFramebuffersEXT(1,&f1.id);
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f1.id);

 // 1 depth
 glGenRenderbuffersEXT(1,&f1.depth);
 glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,f1.depth);
 glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT,scr_w,scr_h);
 glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,f1.depth);
 glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,0);
 
 // 1 frame
 glGenTextures(1,&(f1.tex_id));
 glBindTexture(GL_TEXTURE_2D,f1.tex_id);

 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,scr_w,scr_h,0,GL_RGBA,GL_FLOAT,NULL);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
 glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,f1.tex_id,0);

 ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
 if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
  printf("FBO1 error: 0x%x\n",ret); return(1);
 }

 printf("FBO1 frame textura: %u\n",f1.tex_id);

 // FBO2 pre scr_w2, scr_h2
 glGenFramebuffersEXT(1,&f2.id);
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f2.id);

 // 0 depth
 // 3 frame
 loopi(3) {
  glGenTextures(1,&(f2.tex_id[i]));
  glBindTexture(GL_TEXTURE_2D,f2.tex_id[i]);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,scr_w2,scr_h2,0,GL_RGBA,GL_FLOAT,NULL);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  switch (i) {
   case 0: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,f2.tex_id[i],0); break;
   case 1: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D,f2.tex_id[i],0); break;
   case 2: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT2_EXT,GL_TEXTURE_2D,f2.tex_id[i],0); break;
  }

  printf("FBO2 frame textura: %u\n",f2.tex_id[i]);
 }

 ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
 if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
  printf("FBO2 error: 0x%x\n",ret); return(1);
 }

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

 init_obj(); // inicializacia stvoruholnika

 created=1; // FBO1 a FBO2 boli uspesne vytvorene

 return(0);
}

// inicializacia FBO pre downsampler ***************************************
int CFBO::init_downsampler(void)
{
 int w=scr_w,h=scr_h; // originalna velkost okna

 // 4 FBO pre downsampler kazdy dalsi s polovicnou velkostou
 loopi(4) {
  w/=2; h/=2; // polovicna velkost okna

  // frame buffer object
  glGenFramebuffersEXT(1,&(f_downsampler[i].id));
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f_downsampler[i].id);

  // depth
  glGenRenderbuffersEXT(1,&(f_downsampler[i].depth));
  glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,f_downsampler[i].depth);
  glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT,w,h);
  glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,f_downsampler[i].depth);
  glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,0);

  // frame
  glGenTextures(1,&(f_downsampler[i].tex_id));
  glBindTexture(GL_TEXTURE_2D,f_downsampler[i].tex_id);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,w,h,0,GL_RGBA,GL_FLOAT,NULL);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,f_downsampler[i].tex_id,0);

  ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
  if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
   printf("FBO for downsampler error: 0x%x\n",ret); return(1);
  }

  printf("FBO pre downsampler frame textura %ix%i: %u\n",w,h,f_downsampler[i].tex_id);
 }

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

 init_obj(); // inicializacia stvoruholnika

 created=2; // FBO pre downsampler bol uspesne vytvoreny

 return(0);
}

// inicializacia FBO pre SSAO **********************************************
int CFBO::init_ssao(void)
{
 // FBO pre scr_w, scr_h
 glGenFramebuffersEXT(1,&f_ssao.id);
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f_ssao.id);

 // 1. frame vo forme 2D textury pre ulozenie hlbok
 glGenTextures(1,&f_ssao.depth);
 glBindTexture(GL_TEXTURE_2D,f_ssao.depth);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
 glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT32,scr_w,scr_h,0,GL_DEPTH_COMPONENT,GL_FLOAT,NULL);
 glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_2D,f_ssao.depth,0);

 printf("FBO_SSAO depth textura: %u\n",f_ssao.depth);

 // 2. frame vo forme 2D textury pre ulozenie normal
 glGenTextures(1,&f_ssao.normal);
 glBindTexture(GL_TEXTURE_2D,f_ssao.normal);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,scr_w,scr_h,0,GL_RGBA,GL_FLOAT,NULL);
 glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,f_ssao.normal,0);

 printf("FBO_SSAO normal textura: %u\n",f_ssao.normal);

 // 3. frame vo forme 2D textury pre ulozenie farieb
 glGenTextures(1,&f_ssao.color);
 glBindTexture(GL_TEXTURE_2D,f_ssao.color);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F_ARB,scr_w,scr_h,0,GL_RGBA,GL_FLOAT,NULL);
 glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D,f_ssao.color,0);

 printf("FBO_SSAO color textura: %u\n",f_ssao.color);

 ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
 if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
  printf("FBO_SSAO error: 0x%x\n",ret); return(1);
 }

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

 init_obj(); // inicializacia stvoruholnika

 created=3; // FBO pre SSAO bol uspesne vytvoreny

 return(0);
}

// inicializacia FBO pre shadowmap *****************************************
int CFBO::init_shadowmap(void)
{
 // FBO pre SHADOWMAPSIZE x SHADOWMAPSIZE
 glGenFramebuffersEXT(1,&f_shadowmap.id);
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f_shadowmap.id);

 // 1 frame vo forme 3D textury pre ulozenie hlbok
 // rozmery textury SHADOWMAPSIZE x SHADOWMAPSIZE x 5
 glGenTextures(1,&f_shadowmap.depth);
 glBindTexture(GL_TEXTURE_2D_ARRAY_EXT,f_shadowmap.depth);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_TEXTURE_COMPARE_MODE,GL_COMPARE_R_TO_TEXTURE);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_TEXTURE_COMPARE_FUNC,GL_LEQUAL);
 glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT,GL_DEPTH_TEXTURE_MODE,GL_INTENSITY);
 glTexImage3D(GL_TEXTURE_2D_ARRAY_EXT,0,GL_DEPTH_COMPONENT,SHADOWMAPSIZE,SHADOWMAPSIZE,5,0,GL_DEPTH_COMPONENT,GL_FLOAT,NULL);
 // vyber 0-tej vrstvy
 glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,f_shadowmap.depth,0,0);

 // ziadny draw buffer pre tento konkretny FBO => nevyplnaju sa RGBA color hodnoty
 glDrawBuffers(0,NULL);

 printf("FBO_SHADOWMAP depth textura: %u\n",f_shadowmap.depth);

 ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
 if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
  printf("FBO_SHADOWMAP error: 0x%x\n",ret); return(1);
 }

 // 0-ta vrstva 3D textury
 glBindTexture(GL_TEXTURE_2D_ARRAY_EXT,0);

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

 //init_obj(); // inicializacia stvoruholnika

 created=4; // FBO pre shadowmap bol uspesne vytvoreny

 return(0);
}

// pripojenie L-tej vrstvy 3D textury **************************************
void CFBO::attach_layer(int L)
{
 glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,f_shadowmap.depth,0,L);
}

// inicializacia FBO pre nespracovane udaje ********************************
int CFBO::init_raw(void)
{
 // rendering v msaa_samples-nasobnom rozliseni
 int w=msaa_samples*scr_w;
 int h=msaa_samples*scr_h;

 // FBO pre w, h
 glGenFramebuffersEXT(1,&f_raw.id);
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,f_raw.id);

 // 1. frame - depth_stencil
 glGenRenderbuffersEXT(1,&(f_raw.tex_id[0]));
 glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,f_raw.tex_id[0]);
 glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH24_STENCIL8,w,h);
 glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_STENCIL_ATTACHMENT,GL_RENDERBUFFER_EXT,f_raw.tex_id[0]);
 glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,0);

 // 2-4. frame - albedo, normal, depth
 loopi(3) {
  glGenTextures(1,&(f_raw.tex_id[i+1]));
  glBindTexture(GL_TEXTURE_2D,f_raw.tex_id[i+1]);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  switch (i) {
   case 0: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,f_raw.tex_id[1],0); break;
   case 1: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT1_EXT,GL_TEXTURE_2D,f_raw.tex_id[2],0); break;
   case 2: glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT2_EXT,GL_TEXTURE_2D,f_raw.tex_id[3],0); break;
  }

  printf("FBO pre nespracovane udaje frame %u textura: %u\n",i+1,f_raw.tex_id[i+1]);
 }
 
 glGenTextures(1,&(f_raw.tex_id[4]));
 glBindTexture(GL_TEXTURE_2D,f_raw.tex_id[4]);

 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA16F,w,h,0,GL_RGBA,GL_FLOAT,NULL);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
 glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT3_EXT,GL_TEXTURE_2D,f_raw.tex_id[4],0);
 printf("FBO pre nespracovane udaje frame %u textura: %u\n",4,f_raw.tex_id[4]);

 ret=glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
 if (ret!=GL_FRAMEBUFFER_COMPLETE_EXT) {
  printf("FBO pre nespracovane udaje error: 0x%x\n",ret); return(1);
 }

 // aplikaciou alokovany framebuffer
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

 // stvoruholnik predstavujuci FBO zlozeny z 2 trojuholnikov
 loopi(18) v[i]=0.0f;
 v[0]=-1.0f; v[1]=-1.0f; v[3]=1.0f; v[4]=-1.0f; v[6]=1.0f; v[7]=1.0f;
 v[9]=-1.0f; v[10]=-1.0f; v[12]=1.0f; v[13]=1.0f; v[15]=-1.0f; v[16]=1.0f;
 t[0]=0.0f; t[1]=0.0f; t[2]=1.0f; t[3]=0.0f; t[4]=1.0f; t[5]=1.0f;
 t[6]=0.0f; t[7]=0.0f; t[8]=1.0f; t[9]=1.0f; t[10]=0.0f; t[11]=1.0f;
 // znizenie maximalnych hodnot texturovacich suradnic
 // o msa_samples * delta dvoch susediacich texturovacich suradnic
 float dx,dy;
 if (msaa_samples>1) {
  dx=(1.0f/(float) w)*(float) (msaa_samples-1);
  dy=(1.0f/(float) h)*(float) (msaa_samples-1);
 }
 else { dx=0.0f; dy=0.0f; }
 t[0]=0.0f; t[1]=0.0f; t[2]=1.0f-dx; t[3]=0.0f; t[4]=1.0f-dx; t[5]=1.0f-dy;
 t[6]=0.0f; t[7]=0.0f; t[8]=1.0f-dx; t[9]=1.0f-dy; t[10]=0.0f; t[11]=1.0f-dy;

 // Vertex Array Object
 glGenVertexArrays(1,&vao);
 glBindVertexArray(vao);

 glGenBuffers(1,&v_buf);
 glBindBuffer(GL_ARRAY_BUFFER,v_buf);
 glBufferData(GL_ARRAY_BUFFER,sizeof(v),v,GL_STATIC_DRAW);
 // vrcholy
 glVertexAttribPointer(0,3,GL_FLOAT,GL_FALSE,0,0);
 glGenBuffers(1,&t_buf);
 glBindBuffer(GL_ARRAY_BUFFER,t_buf);
 glBufferData(GL_ARRAY_BUFFER,sizeof(t),t,GL_STATIC_DRAW);
 // texturovacie suradnice
 glVertexAttribPointer(8,2,GL_FLOAT,GL_FALSE,0,0);

 glEnableVertexAttribArray(0);
 glEnableVertexAttribArray(8);

 glBindVertexArray(0);

 created=5; // FBO pre nespracovane udaje bol uspesne vytvoreny

 return(0);
}

// zrusenie FBO ************************************************************
void CFBO::cancel(void)
{
 switch (created) {
  case 0: // FBO0
          glDeleteRenderbuffersEXT(1,&f0.depth);
          loopi(2) {
           glDeleteTextures(1,&(f0.tex_id[i]));
           printf("zrusenie FBO0 frame textury: %u\n",f0.tex_id[i]);
          }
          glDeleteFramebuffersEXT(1,&f0.id);
          break;
  case 1: // FBO1
          glDeleteTextures(1,&(f1.tex_id));
          glDeleteRenderbuffersEXT(1,&f1.depth);
          glDeleteFramebuffersEXT(1,&f1.id);
          printf("zrusenie FBO1 frame textury: %u\n",f1.tex_id);

          // FBO2
          loopi(3) {
           glDeleteTextures(1,&(f2.tex_id[i]));
           printf("zrusenie FBO2 frame textury: %u\n",f2.tex_id[i]);
          }
          glDeleteFramebuffersEXT(1,&f2.id);
          break;
  case 2: // 4 FBO pre downsampler
          loopi(4) {
           glDeleteRenderbuffersEXT(1,&(f_downsampler[i].depth));
           glDeleteTextures(1,&(f_downsampler[i].tex_id));
           printf("zrusenie FBO pre downsampler frame textury: %u\n",f_downsampler[i].tex_id);
           glDeleteFramebuffersEXT(1,&(f_downsampler[i].id));
          }
          break;
  case 3: // FBO pre SSAO
          glDeleteTextures(1,&f_ssao.depth);
          printf("zrusenie FBO pre SSAO depth textury: %u\n",f_ssao.depth);
          glDeleteTextures(1,&f_ssao.normal);
          printf("zrusenie FBO pre SSAO normal textury: %u\n",f_ssao.normal);
          glDeleteTextures(1,&f_ssao.color);
          printf("zrusenie FBO pre SSAO color textury: %u\n",f_ssao.color);
          glDeleteFramebuffersEXT(1,&f_ssao.id);
          break;
  case 4: // FBO pre shadowmap
          glDeleteTextures(1,&f_shadowmap.depth);
          printf("zrusenie FBO pre SHADOWMAP depth textury: %u\n",f_shadowmap.depth);
          glDeleteFramebuffersEXT(1,&f_shadowmap.id);
          break;
  case 5: // FBO pre nespracovane udaje
          glDeleteRenderbuffersEXT(1,&(f_raw.tex_id[0]));
          loopi(4) {
           glDeleteTextures(1,&(f_raw.tex_id[i+1]));
           printf("zrusenie FBO pre nespracovane udaje frame %u textury: %u\n",i+1,f_raw.tex_id[i+1]);
          }
          glDeleteFramebuffersEXT(1,&f_raw.id);
          break;
  default: break;
 }
}

// vykreslenie obsahu FBO1 *************************************************
void CFBO::drawfbo1_attachment(void)
{
 glBindTexture(GL_TEXTURE_2D,f1.tex_id);
 glBindVertexArray(vao);
 glDrawArrays(GL_TRIANGLES,0,6);
 glBindVertexArray(0);
}

// vykreslenie obsahu FBO2 *************************************************
void CFBO::drawfbo2_attachment(int N)
{
 glBindTexture(GL_TEXTURE_2D,f2.tex_id[N]);
 glBindVertexArray(vao);
 glDrawArrays(GL_TRIANGLES,0,6);
 glBindVertexArray(0);
}

// vykreslenie objektu FBO *************************************************
void CFBO::draw(void)
{
 glBindVertexArray(vao);
 glDrawArrays(GL_TRIANGLES,0,6);
 glBindVertexArray(0);
}
