#version 150

layout(triangles_adjacency) in;             // vstup - trojuholniky s doplnkovymi vrcholmi
layout(triangle_strip,max_vertices=15) out; // vystup - pas trojuholnikov

// VERTEX atributy 3 vstupnych vrcholov + 3 doplnkovych vrcholov
in vec4 pos[6]; // poloha vrcholov v EYE space
in vec3 col[6]; // farby vrcholov
in vec2 uv[6];  // texturovacie suradnice prisluchajuce vrcholom

// bude sa kreslit trojuholnik / stvoruholnik
flat out bool EdgeQuad;

// vystupne udaje pre fragment shader
out block
{
 vec3 nor;
 vec2 uv;
 vec3 dir_to_light,dir_to_eye;
 float dist;
 vec4 ambient,diffuse,specular;
} forFS;

// poloha vrcholov v EYE a CLIP space
vec4 eyepos[3],clippos[3];
// pomocne vektory a normala
vec4 v[2],n,dte[3];
vec2 w,vx,vy;

// vypocet normaly trojuholnika v EYE priestore
void compute_normal(void)
{
 v[0]=eyepos[1]-eyepos[0]; // vektor 0 -> 1
 v[1]=eyepos[2]-eyepos[0]; // vektor 0 -> 2
 n=vec4(cross(v[0].xyz,v[1].xyz),0.0); // vektorovy sucin=normala
 n=normalize(n);                       // normalizacia normaly
}

// vykreslenie stvoruholnika
void EmitEdgeQuad(void)
{
 // 2D vektor v[0] -> v[1]
 w.xy=normalize(v[1].xy-v[0].xy);
 vx=w.xy*0.08;           // predlzenie stran stvoruholnika v smere vx
 vy=vec2(-w.y,w.x)*0.15; // vektor kolmy na vx (predlzenie)

 EdgeQuad=true; // bude sa kreslit stvoruholnik

 // pas trojuholnikov = stvoruholnik
 gl_Position=vec4(v[0].xy-vx,   v[0].z,v[0].w); EmitVertex();
 gl_Position=vec4(v[0].xy-vx-vy,v[0].z,v[0].w); EmitVertex();
 gl_Position=vec4(v[1].xy+vx,   v[1].z,v[1].w); EmitVertex();
 gl_Position=vec4(v[1].xy+vx-vy,v[1].z,v[1].w); EmitVertex();
 EndPrimitive();
}

void main(void)
{
 // poloha vrcholov trojuholnika v CLIP priestore
 clippos[0]=gl_in[0].gl_Position;
 clippos[1]=gl_in[2].gl_Position;
 clippos[2]=gl_in[4].gl_Position;

 // poloha vrcholov trojuholnika v EYE priestore
 eyepos[0]=pos[0];
 eyepos[1]=pos[2];
 eyepos[2]=pos[4];
 compute_normal(); // normala v EYE priestore
 dte[0]=normalize(-pos[0]); // vektor VERTEX0 -> EYE v EYE priestore

 // ak je trojuholnik viditelny (privrateny k oku pozorovatela)
 if (dot(n,dte[0])>0.0) {

  EdgeQuad=false; // bude sa kreslit trojuholnik

  // 1. vrchol
  // poloha vrcholu v CLIP priestore
  gl_Position=clippos[0];
  // normala
  forFS.nor=n.xyz;
  // texturovacie suradnice
  forFS.uv=uv[0];
  // vektor VERTEX -> LIGHT v EYE space
  forFS.dir_to_light=(gl_LightSource[0].position.xyz-eyepos[0].xyz);
  forFS.dist=length(forFS.dir_to_light);
  // vektor VERTEX0 -> EYE v EYE space
  forFS.dir_to_eye=dte[0].xyz;
  // ambient
  forFS.ambient=(gl_FrontLightModelProduct.sceneColor+
                 gl_FrontLightProduct[0].ambient)*vec4(col[0],1.0);
  // diffuse
  forFS.diffuse=gl_FrontLightProduct[0].diffuse;
  // specular
  forFS.specular=gl_FrontLightProduct[0].specular;
  EmitVertex();

  // 2. vrchol
  // poloha vrcholu v CLIP priestore
  gl_Position=clippos[1];
  // normala
  forFS.nor=n.xyz;
  // texturovacie suradnice
  forFS.uv=uv[2];
  // vektor VERTEX -> LIGHT v EYE space
  forFS.dir_to_light=(gl_LightSource[0].position.xyz-eyepos[1].xyz);
  forFS.dist=length(forFS.dir_to_light);
  // vektor VERTEX1 -> EYE v EYE space
  dte[1]=normalize(-pos[2]);
  forFS.dir_to_eye=dte[1].xyz;
  // ambient
  forFS.ambient=(gl_FrontLightModelProduct.sceneColor+
                 gl_FrontLightProduct[0].ambient)*vec4(col[2],1.0);
  // diffuse
  forFS.diffuse=gl_FrontLightProduct[0].diffuse;
  // specular
  forFS.specular=gl_FrontLightProduct[0].specular;
  EmitVertex();

  // 3. vrchol
  // poloha vrcholu v CLIP priestore
  gl_Position=clippos[2];
  // normala
  forFS.nor=n.xyz;
  // texturovacie suradnice
  forFS.uv=uv[4];
  // vektor VERTEX -> LIGHT v EYE space
  forFS.dir_to_light=(gl_LightSource[0].position.xyz-eyepos[2].xyz);
  forFS.dist=length(forFS.dir_to_light);
  // vektor VERTEX2 -> EYE v EYE space
  dte[2]=normalize(-pos[4]);
  forFS.dir_to_eye=dte[2].xyz;
  // ambient
  forFS.ambient=(gl_FrontLightModelProduct.sceneColor+
                 gl_FrontLightProduct[0].ambient)*vec4(col[4],1.0);
  // diffuse
  forFS.diffuse=gl_FrontLightProduct[0].diffuse;
  // specular
  forFS.specular=gl_FrontLightProduct[0].specular;
  EmitVertex();
  EndPrimitive();

  // test viditelnosti doplkoveho trojuholnika priliehajuceho k 1. strane
  eyepos[0]=pos[0]; eyepos[1]=pos[1]; eyepos[2]=pos[2];
  compute_normal();
  if (dot(n,dte[1])<0.0) {
   v[0]=clippos[0]; v[1]=clippos[1]; EmitEdgeQuad();
  }
  // test viditelnosti doplnkoveho trojuholnika priliehajuceho k 2. strane
  eyepos[0]=pos[2]; eyepos[1]=pos[3]; eyepos[2]=pos[4];
  compute_normal();
  if (dot(n,dte[2])<0.0) {
   v[0]=clippos[1]; v[1]=clippos[2]; EmitEdgeQuad();
  }
  // test viditelnosti doplnkoveho trojuholnika priliehajuceho k 3. strane
  eyepos[0]=pos[4]; eyepos[1]=pos[5]; eyepos[2]=pos[0];
  compute_normal();
  if (dot(n,dte[0])<0.0) {
   v[0]=clippos[2]; v[1]=clippos[0]; EmitEdgeQuad();
  }
 }
}
