#version 150

layout(triangles) in;
layout(triangle_strip,max_vertices=3) out;

uniform mat4 modelview_mat;  // modelview matica
uniform mat4 projection_mat; // projection matica
uniform float sl;            // uroven zuzenia
uniform float ra;            // uhol rotacie

// VERTEX atributy 3 vstupnych vrcholov
in vec3 col[3]; // farby vrcholov
in vec2 uv[3];  // texturovacie suradnice prisluchajuce vrcholom

// vystupne udaje pre fragment shader
out block
{
 vec3 nor;
 vec2 uv;
 vec3 dir_to_light,dir_to_eye;
 float dist;
 vec4 ambient,diffuse,specular;
} forFS;

// poloha vrcholov v OBJECT, EYE a CLIP space
vec4 objpos[3],eyepos[3],clippos[3];
// tazisko trojuholnika
vec4 s=vec4(0.0,0.0,0.0,1.0);
// pomocne vektory a normala
vec4 v[2],n;
// pomocne premenne a matica
float co,co1,si;
mat4 m;

// vypocet normaly trojuholnika
void compute_normal(void)
{
 v[0]=objpos[1]-objpos[0]; // vektor 0 -> 1
 v[1]=objpos[2]-objpos[0]; // vektor 0 -> 2
 n=vec4(cross(v[0].xyz,v[1].xyz),0.0); // vektorovy sucin=normala
 n=normalize(n);                       // normalizacia normaly
}

// rotacia o uhol A okolo osi X Y Z - os=vektor musi byt normalizovany
void rotate(float A,float X,float Y,float Z) {
 // naplnenie pomocnej matice
 co=cos(A); co1=1.0-co; si=sin(A);
 m[0]=vec4(X*X*co1+co,X*Y*co1+Z*si,X*Z*co1-Y*si,0.0);
 m[1]=vec4(Y*X*co1-Z*si,Y*Y*co1+co,Y*Z*co1+X*si,0.0);
 m[2]=vec4(Z*X*co1+Y*si,Z*Y*co1-X*si,Z*Z*co1+co,0.0);
 m[3]=vec4(0.0,0.0,0.0,1.0);

 // posun trojuholnika (taziska) do stredu suradnicovej sustavy
 objpos[0]-=s; objpos[1]-=s; objpos[2]-=s;

 // rotacia trojuholnika (jeho 3 vrcholov)
 objpos[0]=m*objpos[0];
 objpos[1]=m*objpos[1];
 objpos[2]=m*objpos[2];

 // posun trojuholnika (taziska) na povodnu polohu
 objpos[0]+=s; objpos[1]+=s; objpos[2]+=s;
}

// vypocet polohy vrcholu (vratane zuzenia)
void compute_position(int I)
{
 // poloha vrcholu v EYE space
 eyepos[I]=modelview_mat*(s+sl*(objpos[I]-s));

 // poloha vrcholu v CLIP space
 clippos[I]=projection_mat*eyepos[I];

 gl_Position=clippos[I];
}

void main(void)
{
 int i;

 // poloha 3 vrcholov v OBJECT space
 objpos[0]=gl_in[0].gl_Position;
 objpos[1]=gl_in[1].gl_Position;
 objpos[2]=gl_in[2].gl_Position;

 // tazisko trojuholnika
 s.xyz=(objpos[0].xyz+objpos[1].xyz+objpos[2].xyz)/3.0;

 // vypocet normaly trojuholnika
 compute_normal();

 // rotacia trojuholnika okolo normaly vo vlastnom tazisku
 // poloha vrcholov sa zmeni, avsak normala nie,
 // normalu nie je nutne znovu vypocitat
 rotate(ra,n.x,n.y,n.z);

 // pre 3 vrcholy
 for (i=0;i<3;i++) {
  // vypocet polohy vrcholu v CLIP space (vratane zuzenia)
  compute_position(i);

  // normala
  //nor=(gl_NormalMatrix*in_nor);
  // ak sa pouziva rovnomerne skalovanie,
  // potom gl_NormalMatrix=gl_ModelViewMatrix
  // predpoklada sa, ze normala je normalizovana
  forFS.nor=vec3(modelview_mat*n);

  // texturovacie suradnice
  forFS.uv=uv[i];

  // vektor VERTEX -> LIGHT v EYE space
  forFS.dir_to_light=(gl_LightSource[0].position.xyz-eyepos[i].xyz);
  forFS.dist=length(forFS.dir_to_light);
  // vektor VERTEX -> EYE v EYE space
  forFS.dir_to_eye=(-eyepos[i].xyz);

  // ambient
  forFS.ambient=(gl_FrontLightModelProduct.sceneColor+
                 gl_FrontLightProduct[0].ambient)*vec4(col[i],1.0);
  // diffuse
  forFS.diffuse=gl_FrontLightProduct[0].diffuse;
  // specular
  forFS.specular=gl_FrontLightProduct[0].specular;

  EmitVertex();
 }

 EndPrimitive();
}
